# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/05/2021"


from est import settings
import silx.io.h5py_utils
import silx.io.utils
import os


@silx.io.h5py_utils.retry(retry_timeout=settings.DEFAULT_READ_TIMEOUT)
def get_data(url):
    """Returns a numpy data from an URL.

    Examples:

    >>> # 1st frame from an EDF using silx.io.open
    >>> data = silx.io.get_data("silx:/users/foo/image.edf::/scan_0/instrument/detector_0/data[0]")

    >>> # 1st frame from an EDF using fabio
    >>> data = silx.io.get_data("fabio:/users/foo/image.edf::[0]")

    Yet 2 schemes are supported by the function.

    - If `silx` scheme is used, the file is opened using
        :meth:`silx.io.open`
        and the data is reach using usually NeXus paths.
    - If `fabio` scheme is used, the file is opened using :meth:`fabio.open`
        from the FabIO library.
        No data path have to be specified, but each frames can be accessed
        using the data slicing.
        This shortcut of :meth:`silx.io.open` allow to have a faster access to
        the data.

    .. seealso:: :class:`silx.io.url.DataUrl`

    :param Union[str,silx.io.url.DataUrl]: A data URL
    :rtype: Union[numpy.ndarray, numpy.generic]
    :raises ImportError: If the mandatory library to read the file is not
        available.
    :raises ValueError: If the URL is not valid or do not match the data
    :raises IOError: If the file is not found or in case of internal error of
        :meth:`fabio.open` or :meth:`silx.io.open`. In this last case more
        informations are displayed in debug mode.
    """
    if not isinstance(url, silx.io.url.DataUrl):
        url = silx.io.url.DataUrl(url)

    if not url.is_valid():
        raise ValueError("URL '%s' is not valid" % url.path())

    if not os.path.exists(url.file_path()):
        raise IOError("File '%s' not found" % url.file_path())

    if url.scheme() == "silx":
        data_path = url.data_path()
        data_slice = url.data_slice()

        with silx.io.h5py_utils.File(url.file_path(), "r") as h5:
            if data_path not in h5:
                raise ValueError("Data path from URL '%s' not found" % url.path())
            data = h5[data_path]

            if not silx.io.is_dataset(data):
                raise ValueError(
                    "Data path from URL '%s' is not a dataset" % url.path()
                )

            if data_slice is not None:
                data = silx.io.utils.h5py_read_dataset(data, index=data_slice)
            else:
                # works for scalar and array
                data = silx.io.utils.h5py_read_dataset(data)

    elif url.scheme() == "fabio":
        import fabio

        data_slice = url.data_slice()
        if data_slice is None:
            data_slice = (0,)
        if data_slice is None or len(data_slice) != 1:
            raise ValueError(
                "Fabio slice expect a single frame, but %s found" % data_slice
            )
        index = data_slice[0]
        if not isinstance(index, int):
            raise ValueError(
                "Fabio slice expect a single integer, but %s found" % data_slice
            )

        try:
            fabio_file = fabio.open(url.file_path())
        except Exception:
            raise IOError(
                "Error while opening %s with fabio (use debug for more information)"
                % url.path()
            )

        if fabio_file.nframes == 1:
            if index != 0:
                raise ValueError(
                    "Only a single frame available. Slice %s out of range" % index
                )
            data = fabio_file.data
        else:
            data = fabio_file.getframe(index).data

        # There is no explicit close
        fabio_file = None

    else:
        raise ValueError("Scheme '%s' not supported" % url.scheme())

    return data



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/12/2020"


from silx.io.url import DataUrl
from silx.io.h5py_utils import File as HDF5File
import est.io
import copy
import numpy
import logging
import h5py
import tempfile
import os
import shutil
import pint
from est.units import ur
from est.units import convert_to as convert_unit_to
from silx.io.utils import h5py_read_dataset
from typing import Union
from .dim import Dim
from .spectra import Spectra
from est.settings import DEFAULT_READ_TIMEOUT

try:
    import larch
except ImportError:
    _has_larch = False
else:
    _has_larch = True

_logger = logging.getLogger(__name__)


class XASObject(object):
    """Base class of XAS

    :param spectra: absorbed beam as a list of :class:`.Spectrum` or a
                    numpy.ndarray. If is a numpy array:
                        * dim0: channel,
                        * dim1: Y,
                        * dim2: X
    :type: Union[numpy.ndarray, list]
    :param energy: beam energy
    :type: numpy.ndarray of one dimension
    :param dict configuration: configuration of the different process
    :param int dim1: first dimension of the spectra
    :param int dim2: second dimension of the spectra
    :param str name: name of the object. Will be used for the hdf5 entry
    :param bool keep_process_flow: if True then will keep the trace of the set
                                   of process applied to the XASObject into a
                                   hdf5 file.
                                   This is also used for the
    :param spectra_url: path to the spectra data if any. Used when serializing
                        the XASObject. Won't read it from it.
    :type: Union[None,str]
    :param spectra_url_dimensions: dimensions of the stored spectra. WARNING:
                                   this is different of the spectra dimension
                                   which should be given in the
                                   `DEFAULT_DIMENSIONS` order (Channel, Y, X)
    :type: Union[tuple,None]
    :param energy_url: path to the energy / channel  data if any. Used when
                       serializing the XASObject. Won't read it from it.
    :type: Union[None,str]
    """

    DEFAULT_DIMENSIONS = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)

    def __init__(
        self,
        spectra=None,
        energy=None,
        configuration: Union[dict, None] = None,
        dim1: Union[int, None] = None,
        dim2: Union[int, None] = None,
        name: str = "scan1",
        keep_process_flow: bool = True,
        spectra_url: Union[DataUrl, None] = None,
        spectra_url_dims=None,
        energy_url: Union[DataUrl, None] = None,
        check_energy: bool = True,
    ):
        if isinstance(energy, numpy.ndarray) and not isinstance(
            energy, pint.quantity.Quantity
        ):
            # automatic energy unit guess. convert energy provided in keV to eV
            # works because EXAFS data are never longer than 2 keV and a low
            # energy XANES is never shorter than 3 eV. See issue 30
            if (energy.max() - energy.min()) < 3.0:
                energy = energy * ur.keV
            else:
                energy = energy * ur.eV

        if energy is not None:
            energy = energy.m_as(ur.eV)

        self.__channels = None
        self.__spectra = Spectra(energy=energy)
        self.__dim1 = None
        self.__dim2 = None
        self.__processing_index = 0
        self.__h5_file = None
        self.__entry_name = name
        self.__spectra_url = spectra_url
        self.__spectra_url_dims = spectra_url_dims
        self.__energy_url = energy_url
        self.spectra = (energy, spectra, dim1, dim2)
        self.configuration = configuration
        if keep_process_flow is True:
            self.__h5_file = os.path.join(
                tempfile.mkdtemp(), "_".join((name, "_flow.h5"))
            )
            self.link_to_h5(self.__h5_file)

    def attach_3d_array(self, attr_name, attr_value):
        """
        Attach to each attribute the provided value

        :param attr_name:
        :param attr_value:
        :return:
        """
        if isinstance(attr_value, DataUrl):
            data = get_data(attr_value)
        else:
            data = attr_value
        if not isinstance(data, numpy.ndarray):
            raise TypeError(
                "`attr_value` should be an instance of DataUrl "
                "referencing a numpy array or a numpy array not "
                "{}".format(type(data))
            )
        if data.ndim > 2:
            for spectrum, d in zip(self.spectra.data.flat, data):
                setattr(spectrum, attr_name, d)
        else:
            for spectrum in self.spectra.data.flat:
                setattr(spectrum, attr_name, data)

    @property
    def entry(self) -> str:
        return self.__entry_name

    @property
    def spectra(self):
        return self.__spectra

    @property
    def spectra_url(self) -> Union[None, DataUrl]:
        """Url from where the spectra is available.
        Used for object serialization"""
        return self.__spectra_url

    @spectra_url.setter
    def spectra_url(self, url: DataUrl):
        self.__spectra_url = url

    @property
    def spectra_url_dims(self) -> tuple:
        """used to interpret the spectra_url if any"""
        return self.__spectra_url_dims

    @spectra_url_dims.setter
    def spectra_url_dims(self, dims: tuple):
        self.__spectra_url_dims = dims

    @property
    def energy_url(self) -> Union[DataUrl, None]:
        """Url from where the energy is available.
        Used for object serialization"""
        return self.__energy_url

    @energy_url.setter
    def energy_url(self, url: DataUrl):
        self.__energy_url = url

    @property
    def normalized_energy(self):
        if self.spectra is not None:
            return self.spectra.data.flat[0].normalized_energy

    # TODO: should no take such a tuple in parameter but the different parameter
    # energy, spectra, dim1, dim2
    @spectra.setter
    def spectra(self, energy_spectra: tuple):
        if isinstance(energy_spectra, Spectra):
            self.__spectra = energy_spectra
        else:
            if not len(energy_spectra) == 4:
                raise ValueError(
                    "4 elements expected: energy, spectra, dim1 and dim2. get {} instead".format(
                        len(energy_spectra)
                    )
                )
            energy, spectra, dim1, dim2 = energy_spectra
            self.__spectra = Spectra(energy=energy, spectra=spectra)
            if dim1 is not None and dim2 is not None:
                self.__spectra.reshape((dim1, dim2))

    def get_spectrum(self, dim1_idx: int, dim2_idx: int) -> spectra:
        """Util function to access the spectrum at dim1_idx, dim2_idx"""
        return self.spectra[dim1_idx, dim2_idx]

    @property
    def energy(self) -> numpy.ndarray:
        """energy as a numpy array in eV.
        :note: cannot be with unit because use directly by xraylarch and pymca
        """
        return self.spectra.energy

    @property
    def configuration(self) -> dict:
        return self.__configuration

    @configuration.setter
    def configuration(self, configuration: dict):
        assert configuration is None or isinstance(configuration, dict)
        self.__configuration = configuration or {}

    def _create_saving_pt(self):
        if not self.has_linked_file():
            raise ValueError(
                "there is not where to create a saving pt"
                "(no h5 file linked to the XASObject)"
            )
        else:

            def get_list_spectra():
                res = {}

                for i_spectrum, spectrum in enumerate(self.spectra.data.flat):
                    res[i_spectrum] = spectrum.to_dict()
                return res

            from est.io import write_spectrum_saving_pt

            # save spectra
            entry = "/".join((self.entry, "est_saving_pt", "spectra"))
            for i_spectrum, spectrum in get_list_spectra().items():
                path = "/".join((entry, str(i_spectrum)))
                write_spectrum_saving_pt(
                    h5_file=self.linked_h5_file,
                    entry=path,
                    obj=spectrum,
                    overwrite=True,
                )
            # save channel
            with HDF5File(self.linked_h5_file, "a") as h5f:
                entry = "/".join((self.entry, "est_saving_pt", "channel"))
                if entry in h5f:
                    del h5f[entry]
                h5f[entry] = (
                    "None"
                    if self.energy is None
                    else convert_unit_to(self.energy, ur.eV).m_as(ur.eV)
                )

    def to_dict(self, with_process_details=True) -> dict:
        """convert the XAS object to a dict

        By default made to simply import raw data.

        :param with_process_details: used to embed a list of spectrum with
                                     intermediary result instead of only raw mu.
                                     This is needed especially for the
                                     pushworkflow actors to keep a trace of the
                                     processes.
        :type: bool
        """

        def get_spectra_and_processing():
            if not self.has_linked_file():
                raise ValueError(
                    "To get process details you should have a" "`process` link file"
                )
            else:
                # store the current spectra with processing information
                data_path = "/".join((self.entry, "est_saving_pt", "spectra"))
                return DataUrl(
                    file_path=self.linked_h5_file, data_path=data_path, scheme="est"
                ).path()

        def get_energy():
            if not self.has_linked_file():
                raise ValueError(
                    "To get process details you should have a" "`process` link file"
                )
            else:
                data_path = "/".join((self.entry, "est_saving_pt", "channel"))
                return DataUrl(
                    file_path=self.linked_h5_file, data_path=data_path, scheme="silx"
                ).path()

        self._create_saving_pt()
        spectra_ = get_spectra_and_processing()
        res = {
            "configuration": self.configuration,
            "spectra": spectra_,
            "energy": get_energy(),
            "dim1": self.spectra.shape[0],
            "dim2": self.spectra.shape[1],
        }
        if with_process_details is True:
            res["linked_h5_file"] = self.linked_h5_file
            res["current_processing_index"] = self.__processing_index

        return res

    def absorbed_beam(self) -> numpy.ndarray:
        return self.spectra.map_to(data_info="mu")

    def load_frm_dict(self, ddict: dict):
        """load XAS values from a dict"""
        dimensions = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)
        # default yay of storing data
        from est.io import load_data  # avoid cyclic import

        contains_config_spectrum = "configuration" in ddict or "spectra" in ddict
        """The dict can be on the scheme of the to_dict function, containing
        the spectra and the configuration. Otherwise we consider it is simply
        the spectra"""
        if "configuration" in ddict:
            self.configuration = ddict["configuration"]
        if "spectra" in ddict:
            self.spectra = Spectra.frm_dict(ddict["spectra"], dimensions=dimensions)
        else:
            self.spectra = None
        if "energy" in ddict:
            energy = ddict["energy"]
            if isinstance(energy, str):
                energy_url = DataUrl(path=energy)
                energy = load_data(
                    data_url=energy_url, name="energy", dimensions=dimensions
                )
        else:
            energy = None
        if "dim1" in ddict:
            dim1 = ddict["dim1"]
        else:
            dim1 = None
        if "dim2" in ddict:
            dim2 = ddict["dim2"]
        else:
            dim2 = None
        if "linked_h5_file" in ddict:
            assert "current_processing_index" in ddict
            self.link_to_h5(ddict["linked_h5_file"])
            self.__processing_index = ddict["current_processing_index"]
        self.spectra.reshape((dim1, dim2))
        return self

    @staticmethod
    def from_dict(ddict: dict):
        return XASObject().load_frm_dict(ddict=ddict)

    @staticmethod
    def from_file(
        h5_file,
        entry="scan1",
        spectra_path="data/absorbed_beam",
        energy_path="data/energy",
        configuration_path="configuration",
        dimensions=None,
        timeout=DEFAULT_READ_TIMEOUT,
    ):
        # load only mu and energy from the file
        spectra_url = DataUrl(
            file_path=h5_file, data_path="/".join((entry, spectra_path)), scheme="silx"
        )
        energy_url = DataUrl(
            file_path=h5_file, data_path="/".join((entry, energy_path)), scheme="silx"
        )
        if configuration_path is None:
            config_url = None
        else:
            config_url = DataUrl(
                file_path=h5_file,
                data_path="/".join((entry, configuration_path)),
                scheme="silx",
            )
        from est.io.utils.information import InputInformation

        sp, en, conf = est.io.read_xas(
            InputInformation(
                spectra_url=spectra_url,
                channel_url=energy_url,
                config_url=config_url,
                dimensions=dimensions,
            ),
            timeout=timeout,
        )
        return XASObject(spectra=sp, energy=en, configuration=conf)

    def dump(self, h5_file: str):
        """dump the XAS object to a file_path within the Nexus format"""
        from est.core.io import XASWriter  # avoid cyclic import

        writer = XASWriter()
        # dump raw data
        writer.output_file = h5_file
        writer.dump_xas(xas_obj=self, write_process=True)

    def copy(self, create_h5_file=False):
        """

        :param bool create_h5_file: True if we want to duplicate the h5 file
                                 This is the case for orange widget for example
                                 where some reprocessing can append and each
                                 process need to keep a clear history of
                                 processes, with no knowledge of next
                                 processing.
        """
        # To have dedicated h5 file we have to create one new h5 file for
        # for each process. For now there is no way to do it differently
        res = copy.copy(self)
        if create_h5_file is True:
            new_h5_file = os.path.join(
                tempfile.mkdtemp(), "_".join((self.__entry_name, "_flow.h5"))
            )
            if self.has_linked_file() and os.path.exists(self.linked_h5_file):
                shutil.copy(src=self.linked_h5_file, dst=new_h5_file)
            res.link_to_h5(new_h5_file)
        return res

    def __eq__(self, other):
        return (
            isinstance(other, XASObject)
            and (
                (self.energy is None and other.energy is None)
                or numpy.array_equal(self.energy, other.energy)  # noqa
            )  # noqa
            and self.configuration == other.configuration  # noqa
            and self.spectra,
            other.spectra,  # noqa
        )

    @property
    def n_spectrum(self) -> int:
        """return the number of spectra"""
        return len(self.spectra.data.flat)

    def spectra_keys(self) -> tuple:
        """keys contained by the spectrum object (energy, mu, normalizedmu...)"""
        return self.spectra.keys()

    @property
    def linked_h5_file(self):
        return self.__h5_file

    def link_to_h5(self, h5_file: str):
        """
        Associate a .h5 file to the XASObject. This can be used for storing
        process flow.

        :param h5_file:
        :return:
        """
        self.__h5_file = h5_file

    def has_linked_file(self) -> bool:
        return self.__h5_file is not None

    def get_next_processing_index(self) -> int:
        self.__processing_index += 1
        return self.__processing_index

    def register_processing(self, process, results, plots) -> None:
        """
        Register one process for the current xas object. This require to having
        link a h5file to this object

        :param process: process to be registred
        :type: :class:`.Process`
        :param results: result of the processing. If there is more than one
                     result then a dictionary with the key under which result
                     should be saved and a numpy.ndarray
        :type: Union[numpy.ndarray,dict]
        """
        est.io.write_xas_proc(
            self.linked_h5_file,
            entry=self.__entry_name,
            processing_order=self.get_next_processing_index(),
            process=process,
            results=results,
            plots=plots,
            overwrite=True,
        )

    def get_process_flow(self) -> dict:
        """

        :return: the dict of process information
        :rtype: dict
        """
        if not self.linked_h5_file:
            _logger.warning(
                "process flow is store in the linked .h5 file. If"
                "no link is defined then this information is not"
                "stored"
            )
            return {}
        else:
            recognized_process = est.io.get_xasproc(
                self.linked_h5_file, entry=self.__entry_name
            )
            know_process = (
                "pymca_normalization",
                "pymca_exafs",
                "pymca_ft",
                "pymca_k_weight",
                "larch_autobk",
                "larch_mback",
                "larch_mback_norm",
                "larch_pre_edge",
                "larch_xftf",
                "noise",
                "noise_savgol",
                "roi",
                "energy-roi",
            )

            def filter_recognized_process(process_list):
                res = []
                for process_ in process_list:
                    if (
                        "program" in process_.keys()
                        and h5py_read_dataset(process_["program"])
                        in know_process  # noqa
                    ):
                        res.append(process_)
                return res

            recognized_process = filter_recognized_process(recognized_process)

            def get_ordered_process(process_list):
                res = {}
                for process_ in process_list:
                    if "processing_order" not in process_:
                        _logger.warning(
                            "one processing has not processing order: "
                            + process_["program"]  # noqa
                        )
                    else:
                        processing_order = int(process_["processing_order"])
                        res[processing_order] = process_
                return res

            return get_ordered_process(recognized_process)

    def clean_process_flow(self) -> None:
        """
        Remove existing process flow
        """
        if not self.linked_h5_file:
            _logger.warning(
                "process flow is store in the linked .h5 file. If"
                "no link is defined then this information is not"
                "stored"
            )
        else:
            process_flow = self.get_process_flow()
            with HDF5File(self.linked_h5_file, "a") as h5f:
                for index, process_ in process_flow.items():
                    del h5f[process_["_h5py_path"]]

    def copy_process_flow_to(self, h5_file_target: str) -> None:
        """
        copy all the recognized process from self.__h5_file to h5_file_target

        :param str h5_file_target: path to the targeted file. Should be an
                                   existing hdf5 file.
        """
        assert os.path.exists(h5_file_target)
        assert h5py.is_hdf5(h5_file_target)

        flow = self.get_process_flow()
        entry = self.entry
        with HDF5File(self.__h5_file, "a") as source_hdf:
            with HDF5File(h5_file_target, "a") as target_hdf:
                target_entry = target_hdf.require_group(entry)

                def remove_entry_prefix(name):
                    return name.replace("/" + entry + "/", "", 1)

                for process_id, process in flow.items():
                    process_path = process["_h5py_path"]
                    dst_path = remove_entry_prefix(name=process_path)
                    target_entry.copy(source=source_hdf[process_path], dest=dst_path)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""wrapper to the larch mback process"""

from est.core.types import Spectrum, XASObject
from est.core.process.process import Process
from est.core.process.process import _NexusDatasetDef
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from larch.xafs.mback import mback_norm
from larch.xafs.pre_edge import preedge
import multiprocessing
import functools
import logging

_logger = logging.getLogger(__name__)

_DEBUG = True
if _DEBUG:
    from larch.symboltable import Group


def process_spectr_mback_norm(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callback: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug("start mback_norm on spectrum (%s, %s)" % (spectrum.x, spectrum.y))
    assert isinstance(spectrum, Spectrum)
    if not hasattr(spectrum, "norm"):
        _logger.error(
            "spectrum doesn't have norm. Maybe you meed to compute "
            "pre_edge first? Unable to compute mback_norm."
        )
        return None, None
    if not hasattr(spectrum, "pre_edge"):
        _logger.error(
            "spectrum doesn't have norm. Maybe you meed to compute "
            "pre_edge first? Unable to compute mback_norm."
        )
        return None, None

    _conf = configuration
    if "mback_norm" in _conf:
        _conf = _conf["mback_norm"]
    opts = {}

    if _DEBUG is True:
        assert isinstance(spectrum, Group)
    if overwrite:
        _spectrum = spectrum
    else:
        _spectrum = Spectrum().load_frm_dict(spectrum.to_dict())
    # TODO: computing each time preedge should be avoidable
    pre_edge_details = preedge(_spectrum.energy, _spectrum.mu)
    for opt_name in (
        "z",
        "edge",
        "e0",
        "pre1",
        "pre2",
        "norm1",
        "norm2",
        "nnorm",
        "nvict",
    ):
        if opt_name in _conf:
            opts[opt_name] = _conf[opt_name]
        elif pre_edge_details is not None and opt_name in pre_edge_details:
            opts[opt_name] = pre_edge_details[opt_name]

    mback_norm(_spectrum, group=_spectrum, **opts)
    if callbacks:
        for callback in callbacks:
            callback()
    return configuration, _spectrum


def larch_mback_norm(xas_obj):
    """

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[XASObject, dict]
    :return: spectra dict
    :rtype: XASObject
    """
    mback_obj = Larch_mback_norm()
    return mback_obj.process(xas_obj=xas_obj)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


class Larch_mback_norm(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="mback_norm")

    def set_properties(self, properties):
        if "_larchSettings" in properties:
            self._settings = properties["_larchSettings"]

    def process(self, xas_obj):
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        if self._settings:
            _xas_obj.configuration["mback_norm"] = self._settings

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        data_keys = [
            _NexusDatasetDef("mback_mu"),
            _NexusDatasetDef("norm_mback"),
        ]
        if _xas_obj.n_spectrum > 0 and hasattr(_xas_obj.spectra[0], "edge_step"):
            data_keys += ["norm"]
        self.register_process(_xas_obj, data_keys=data_keys)
        return _xas_obj

    def _pool_process(self, xas_obj):
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            for spectrum in xas_obj.spectra:
                process_spectr_mback_norm(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
        else:
            from multiprocessing import Manager

            manager = Manager()
            output_dict = {}
            res_list = manager.list()
            for i_spect, spect in enumerate(xas_obj.spectra):
                res_list.append(None)
                output_dict[spect] = i_spect

            with multiprocessing.Pool(5) as p:
                partial_ = functools.partial(
                    process_spectr_mback_norm,
                    configuration=xas_obj.configuration,
                    callback=self._advancement.increaseAdvancement,
                    overwrite=False,
                    output=res_list,
                    output_dict=output_dict,
                )
                p.map(partial_, xas_obj.spectra)

            # then update local spectrum
            for spectrum, res in zip(xas_obj.spectra, res_list):
                spectrum.update(res)

    def definition(self):
        return "mback norm calculation"

    def program_version(self):
        import larch.version

        return larch.version.version_data()["larch"]

    def program_name(self):
        return "larch_mback_norm"

    __call__ = process



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/25/2019"

import os
import unittest
import tempfile
import shutil
import urllib.request
import numpy
from est.core.types import Spectrum

try:
    import larch
except ImportError:
    has_larch = False
else:
    has_larch = True
    from est.core.process.larch.xftf import larch_xftf, process_spectr_xftf as xftf
    from est.core.process.larch.autobk import process_spectr_autobk as autobk
    from est.core.process.larch.pre_edge import process_spectr_pre_edge as pre_edge
    from est.io.utils.larch import read_ascii


@unittest.skipIf(has_larch is False, "xraylarch not installed")
class TestLarchWorkflows(unittest.TestCase):
    """Test several larch workflow"""

    def setUp(self):
        self.outputdir = tempfile.mkdtemp()
        xmu_url = "https://raw.githubusercontent.com/xraypy/xraylarch/master/examples/xafs/cu_rt01.xmu"
        self.data_file = os.path.join(self.outputdir, "cu_rt01.xmu")

        with urllib.request.urlopen(xmu_url) as response, open(
            self.data_file, "wb"
        ) as out_file:
            data = response.read()  # a `bytes` object
            out_file.write(data)

        assert os.path.exists(self.data_file)
        energy, mu = read_ascii(self.data_file)
        self.spectrum = Spectrum(energy=energy, mu=mu)

    def tearDown(self):
        shutil.rmtree(self.outputdir)

    def testAutobk(self):
        """equivalent treatment as the 'xafs.autobk.par script from larch'"""
        pre_edge(
            spectrum=self.spectrum,
            overwrite=True,
            configuration={"rbkg": 1.0, "kweight": 2},
        )
        autobk(
            spectrum=self.spectrum,
            overwrite=True,
            configuration={
                "kmin": 2,
                "kmax": 16,
                "dk": 3,
                "window": "hanning",
                "kweight": 2,
            },
        )
        xftf(spectrum=self.spectrum, overwrite=True, configuration={"kweight": 2})

    def testXafsft1(self):
        """equivalent treatment as the 'xafs.doc_xafs1.par script from larch'"""
        autobk(
            spectrum=self.spectrum,
            overwrite=True,
            configuration={"rbkg": 1.0, "kweight": 2, "clamp+hi": 10},
        )
        conf, spec_dk1 = xftf(
            spectrum=self.spectrum,
            overwrite=False,
            configuration={
                "kweight": 2,
                "kmin": 3,
                "kmax": 13,
                "window": "hanning",
                "dk": 1,
            },
        )
        conf, spec_dk2 = xftf(
            spectrum=self.spectrum,
            overwrite=False,
            configuration={
                "kweight": 2,
                "kmin": 3,
                "kmax": 13,
                "window": "hanning",
                "dk": 5,
            },
        )
        self.assertNotEqual(spec_dk1, spec_dk2)
        self.assertFalse(numpy.array_equal(spec_dk1.chir_re, spec_dk2.chir_re))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestLarchWorkflows,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")
